/*
 * crypted.c - part of procsms application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "conf.h"
#include "crypted.h"
#include "procsms.h"

#ifdef CONFIG_CRYPTED

/* *INDENT-OFF* */

lc_char PROGMEM lc_Crypted[]={
        LC_EN("Crypted SMS")
        LC_CZ("Sifrovane SMS")
        LC_END
};

static lc_char PROGMEM lc_Send_crypted[]={
        LC_EN("Send Crypted")
        LC_CZ("Poslat sifr.")
        LC_END
};

static lc_char PROGMEM lc_Lock[]={
        LC_EN("Lock")
        LC_CZ("Zamknout")
        LC_END
};

static lc_char PROGMEM lc_Unlock[]={
        LC_EN("Unlock")
        LC_CZ("Odemknout")
        LC_END
};

static lc_char PROGMEM lc_Keys[]={
        LC_EN("Keys")
        LC_CZ("Klice")
        LC_END
};

static lc_char PROGMEM lc_New_Key[]={
        LC_EN("New Key")
        LC_CZ("Novy klic")
        LC_END
};

lc_char PROGMEM lc_Default_Crypted[]={
        LC_EN("Encrypted Message")
        LC_CZ("Sifrovana zprava")
        LC_END
};

static lc_char PROGMEM lc_Key_Main[]={
        LC_EN("Main Key")
        LC_CZ("Hlavni klic")
        LC_END
};

/* *INDENT-ON* */

extern SNodeP crypted_n;

u8 crypted_msg_ctx (SCtx * ctx, u8 action);

extern SNodeP crypted_n_keys_view;
extern SNodeP crypted_n_keys_edit;
extern SNodeP crypted_n_keys_delete;

u8 crypted_keys_ctx (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    spider_append_r (ctx, &crypted_n_keys_delete);
    spider_append_r (ctx, &crypted_n_keys_edit);
    spider_append_r (ctx, &crypted_n_keys_view);
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
  }
  return APP_OK;
}

u8 crypted_keys_delete (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 res;
    SNode *n = ctx->parent->f;
    Pers_mem *p = app_data ();

    res = yesno (rw (&n->text), Q_YESNO_DELETE_QMARK, 0);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES)
    {
      SEdge *e = glob_mem->crypted_keys;
      SEdge *f = e;
      Crypt_key *k = rw (&n->p);

      while (e)
      {
	if (rw (&e->t) == n)
	  break;
	f = e;
	e = rw (&e->next);
      }

      if (f == e)
      {
	ww (&p->crypted_keys, rw (&e->next));
	glob_mem->crypted_keys = rw (&e->next);
      }
      else
      {
	ww (&f->next, rw (&e->next));
      }

      efree (rw (&k->msisdn));
      efree (rw (&n->text));
      efree (rw (&n->p));
      efree (n);
      efree (e);

      spider_back (ctx);
    }
    return APP_BACK;
  }
  return APP_OK;
}

u8 crypted_keys_view (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    SNode *n = ctx->parent->f;
    Crypt_key *key = rw (&n->p);
    u8 i;
    u8 *buf = buf_B ();
    u8 *ms = msisdn2str (rw (&key->msisdn), MSISDN_ADN, MEM_R);

    twofish_ecb_decrypt (&key->text, buf + 2, glob_mem->crypt_key);
    buf[0] = STRING_SLV;
    buf[1] = 0;		//7b dcs
    i = display_text (buf, ms);
    free (ms);
    return i;
  }
  return APP_OK;
}

void make_crypt_key (u8 * buf, u8 * res, u8 key_len)
{
  u8 i;
  u8 tmp[16];

  dbsp ("MAKE_CRYPT_KEY ");
  dbch (key_len);
  dbc ('\n');

  bzero (tmp, 16);

  if (key_len > 16)
    key_len = 16;
  for (i = 0; i < key_len; i++)
    tmp[i] = res[i];
  buf[0] = key_len;
  key_len = dcs_78 (tmp, key_len, DCS_8_TO_7);
  for (i = 1; i < 16; i++)
    buf[i] = 16 - key_len;
  for (i = 0; i < key_len; i++)
    buf[i + 1] = tmp[i];

#ifdef DEBUG
  for (i = 0; i < 16; i++)
  {
    dbch (buf[i]);
    dbc (' ');
  }
  dbc ('\n');
#endif
}

u8 crypted_keys_edit (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    SNode *n = ctx->parent->f;
    Crypt_key *key = rw (&n->p);
    u8 *buf = buf_B ();
    u8 *res;
    u8 i;
    u8 key_len;

    i = password (buf, 16, locale (lc_New_Key));
    if (i != APP_OK)
      return i;

    twofish_ecb_encrypt (buf, &key->text, glob_mem->crypt_key);
    return APP_BACK;
  }
  return APP_OK;
}

u8 crypted_keys_main (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 err;
    Twofish_key *key = NULL;
    u8 *res;
    u8 *buf = buf_B ();
    u8 *k = buf;
    u8 *in = buf + 16;
    u8 *out = in + 16;
    Pers_mem *p = app_data ();
    u8 key_len;
    u8 i;
    SEdge *keys;
    SNode *n;
    Crypt_key *ck;

    i = password (k, 16, locale (lc_New_Key));
    if (i != APP_OK)
      return i;

    key = malloc (sizeof (Twofish_key));
    if (key == NULL)
    {
      err = ERR_NO_RAM;
      goto X_ERR;
    }

#ifdef DEBUG
    dbsp ("MAIN_KEY: ");
    for (i = 0; i < 16; i++)
    {
      dbch (k[i]);
      dbc (' ');
    }
    dbc ('\n');
#endif

    err = twofish_setup (k, 16, key);
    if (err != NO_ERROR)
      goto X_ERR;

    xrand (in);
    xrand (in + 8);
    memcpy (&p->main_key_rnd, in, 16);
    twofish_ecb_encrypt (in, &p->main_key_crnd, key);

    keys = glob_mem->crypted_keys;
    while (keys)
    {
      n = rw (&keys->t);
      ck = rw (&n->p);
      twofish_ecb_decrypt (&ck->text, out, glob_mem->crypt_key);
      twofish_ecb_encrypt (out, &ck->text, key);
      keys = rw (&keys->next);
    }

    free (glob_mem->crypt_key);
    glob_mem->crypt_key = key;

    return APP_BACK;

  X_ERR:
    free (key);
    return perror (err);
  }
  return APP_OK;
}

extern SNodeP crypted_n_keys;

u8 crypted_keys_new (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    Crypt_key *key = NULL;
    u8 *key_msisdn = NULL;
    SNode *n = NULL;
    u8 *n_text = NULL;
    SEdge *e = NULL;
    u8 alpha[ALPHA_LEN];
    u8 alpha_len;
    u8 *res;
    u8 i, j;
    u8 *buf = buf_B ();
    u8 key_len;

    key = emalloc (sizeof (Crypt_key));
    if (key == NULL)
      goto X_ERR;

    n = emalloc (sizeof (SNode));
    if (n == NULL)
      goto X_ERR;

    e = emalloc (sizeof (SEdge));
    if (e == NULL)
      goto X_ERR;

    res = msisdn (alpha);
    if (res == ENULL)
    {
      i = APP_END;
      goto X_END;
    }
    if (res == NULL)
    {
      i = APP_BACK;
      goto X_END;
    }

    key_msisdn = msisdncpy (res, MSISDN_ADN, MEM_E);
    if (key_msisdn == NULL)
      goto X_ERR;

    alpha_len = strlen (alpha) + 1;
    n_text = emalloc (alpha_len);
    if (n_text == NULL)
      goto X_ERR;

    memcpy (n_text, alpha, alpha_len);

    ww (&n->p, key);
    ww (&n->text, n_text);
    ww (&n->cb, crypted_keys_ctx);

    i = password (buf, 16, locale (lc_New_Key));
    if (i != APP_OK)
      goto X_END;

    twofish_ecb_encrypt (buf, &key->text, glob_mem->crypt_key);

    ww (&key->msisdn, key_msisdn);
    ww (&e->f, &crypted_n_keys);
    ww (&e->t, n);
    wb (&e->source, MEM_F_P | MEM_T_E);

    {
      Pers_mem *p = app_data ();

      insert_sorted_alpha (e, &p->crypted_keys, alpha);
      glob_mem->crypted_keys = rw (&p->crypted_keys);
    }

    return APP_BACK;
  X_ERR:
    i = APP_OK;
  X_END:
    efree (key);
    efree (key_msisdn);
    efree (e);
    efree (n);
    efree (n_text);

    if (i == APP_OK)
      perror (ERR_NO_EEPROM);
    return i;
  }
  return APP_OK;
}

u8 crypted_send (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    return send_msg (MSG_CRYPTED, NULL, NULL);
  }
  return APP_OK;
}

u8 crypted_msg_reply (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    SNode *n = ctx->parent->f;
    Crypted_msg *msg = n->p;

    return send_msg (MSG_CRYPTED, NULL, msg->msisdn);
  }
  return APP_OK;
}

u8 crypted_lock (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    free (glob_mem->crypt_key);
    glob_mem->crypt_key = NULL;
    return APP_BACK;
  }
  return APP_OK;
}

u8 crypted_unlock (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 err;
    Twofish_key *key = NULL;
    u8 *res;
    u8 *buf = buf_B ();
    u8 *k = buf;
    u8 *in = buf + 16;
    u8 *out = in + 16;
    Pers_mem *p = app_data ();
    u8 key_len;
    u8 i;

    i = password (k, 16, NULL);
    if (i != APP_OK)
      return i;

    key = malloc (sizeof (Twofish_key));
    if (key == NULL)
    {
      err = ERR_NO_RAM;
      goto X_ERR;
    }

#ifdef DEBUG
    dbsp ("UNLOCK_KEY: ");
    for (i = 0; i < 16; i++)
    {
      dbch (k[i]);
      dbc (' ');
    }
    dbc ('\n');
#endif

    err = twofish_setup (k, 16, key);
    if (err != NO_ERROR)
      goto X_ERR;

    twofish_ecb_encrypt (&p->main_key_rnd, out, key);

#ifdef DEBUG
    dbsp ("IN_RND: ");
    for (i = 0; i < 16; i++)
    {
      dbch (in[i]);
      dbc (' ');
    }
    dbc ('\n');
    dbsp ("OUT_CRND: ");
    for (i = 0; i < 16; i++)
    {
      dbch (out[i]);
      dbc (' ');
    }
    dbc ('\n');

#endif

    if (memcmp (out, p->main_key_crnd, 16))
    {
      err = ERR_WRONG_KEY;
      goto X_ERR;
    }

    glob_mem->crypt_key = key;

    return APP_BACK;
  X_ERR:
    free (key);
    return perror (err);
  }
  return APP_OK;
}

u8 decrypt_msg (u8 * s)
{
  Twofish_key *ak = NULL;
  SNode *n;
  SEdge *msisdns;
  Crypt_key *c_key;
  u8 *t;
  u8 i;
  u8 len;
  u8 *buf = malloc (2 * 16);
  u8 *in = buf + 16;

  if (buf != NULL)
  {
    t = tpdu_seek (s, T_SMS_AA);

    if (glob_mem->crypt_key != NULL)
    {
      msisdns = glob_mem->crypted_keys;
      while (msisdns)
      {
	dbsp ("CR_4\n");
	n = rw (&msisdns->t);
	c_key = rw (&n->p);
	if (msisdncmp (t, MSISDN_SMS, rw (&c_key->msisdn), MSISDN_ADN))
	{
	  dbsp ("CR_5\n");
	  // found crypted key

	  twofish_ecb_decrypt (&c_key->text, in, glob_mem->crypt_key);
	  goto X1;
	}

	msisdns = rw (&msisdns->next);
      }
    }
    i = password (in, 16, NULL);
    if (i != APP_OK)
    {
      free (buf);
      return i;
    }
  X1:

    ak = malloc (sizeof (Twofish_key));
    if (ak)
    {
      u8 *msg;
      u8 j;

      t = tpdu_seek (s, T_SMS_UD);
      msg = t + 12;

      twofish_setup (in, 16, ak);
      i = twofish_cbc_decrypt (msg, msg, 8, ak);
      if (i != NO_ERROR)
      {
	free (buf);
	free (ak);
	return perror (i);
      }

      len = t[20];
      if (len > 0xA0)
	len = 0xA0;

      for (i = 0; i < len; i++)
	t[i] = msg[i + 9];

      *(t - 2) = 0;		// 7b dcs
      *(t - 1) = dcs_7b_len (len);

      free (buf);
      free (ak);
      return APP_OK;
    }
    free (buf);
  }

  return perror (ERR_NO_RAM);
}

u8 crypted_msg_view (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 path[6];
    SNode *n = ctx->parent->f;
    SEdge *msisdns;
    Crypted_msg *msg = n->p;
    u8 i;
    u8 *t;
    Crypt_key *c_key;
    u8 *buf = malloc (2 * 16);
    u8 *in = buf + 16;
    u8 len;

    if (glob_mem->crypt_key != NULL)
    {
      msisdns = glob_mem->crypted_keys;
      while (msisdns)
      {
	dbsp ("CR_4\n");
	n = rw (&msisdns->t);
	c_key = rw (&n->p);
	if (msisdncmp
	    (msg->msisdn, MSISDN_SMS, rw (&c_key->msisdn), MSISDN_ADN))
	{
	  dbsp ("CR_5\n");
	  // found crypted key

	  twofish_ecb_decrypt (&c_key->text, in, glob_mem->crypt_key);
	  goto X1;
	}

	msisdns = rw (&msisdns->next);
      }
    }
    i = password (in, 16, NULL);
    if (i != APP_OK)
    {
      free (buf);
      return i;
    }
  X1:
    memcpy (path, ef_sms_path, 6);
    select (path, 3);
    sim (ME_CMD_READ_RECORD, msg->p1, 0x04, 0xb0, buf_B ());
    memcpy (buf_A (), buf_B (), 0xFF);
    select (0, 0);		// uses buf_B()
    memcpy (buf_B (), buf_A (), 0xFF);

    Twofish_key *ak = malloc (sizeof (Twofish_key));

    if (ak)
    {
      u8 *msg;
      u8 j;
      u8 *s = buf_B ();

      s++;
      i = *s;
      s = s + 1 + i;		//skip SMSC

      t = tpdu_seek (s, T_SMS_UD);
      msg = t + 12;

      twofish_setup (in, 16, ak);
      i = twofish_cbc_decrypt (msg, msg, 8, ak);
      if (i == NO_ERROR)
      {
	len = t[20];
	if (len > 0xA0)
	  len = 0xA0;

	for (i = 0; i < len; i++)
	  t[i] = msg[i + 9];

	*(t - 3) = STRING_SLV;
	*(t - 2) = 0;		// 7b dcs
	*(t - 1) = dcs_7b_len (len);	// nr of septets

	//view
	i = display_text (t - 3, NULL);
      }
      else
	i = perror (i);
    }
    else
      i = perror (ERR_NO_RAM);

    free (ak);
    free (buf);
    return i;
  }
  return APP_OK;
}

extern SNodeP crypted_n_msg_reply;
extern SNodeP crypted_n_msg_view;

u8 crypted_msg_ctx (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    spider_append_r (ctx, &crypted_n_msg_reply);
    spider_append_r (ctx, &crypted_n_msg_view);
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
  }
  return APP_OK;
}

u8 crypted_keys (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    ctx->eE = glob_mem->crypted_keys;
  }
  else if (action == APP_LEAVE)
  {
    ctx->eE = NULL;
  }
  return APP_OK;
}

extern SNodeP crypted_n_locked;
extern SNodeP crypted_n_unlocked;
extern SNodeP crypted_n_keys;

u8 crypted_ctx (SCtx * ctx, u8 action)
{
  dbsp (">crypted_ctx\n");
  if (action == APP_ENTER)
  {
    dbsp (">crypted_ctx_enter\n");
    dbsp ("crypted_msgs ");
    dbih (glob_mem->crypted_msgs);
    dbc ('\n');

    ctx->eR = glob_mem->crypted_msgs;
    glob_mem->crypted_msgs_tmp = glob_mem->crypted_msgs;
    dbsp ("ctx->eR ");
    dbih (ctx->eR);
    dbc ('\n');

    if (glob_mem->crypt_key != NULL)
    {
      spider_append_r (ctx, &crypted_n_keys);
      spider_append_r (ctx, &crypted_n_unlocked);
    }
    else
      spider_append_r (ctx, &crypted_n_locked);
    dbsp ("ctx->eR_1 ");
    dbih (ctx->eR);
    dbc ('\n');
  }
  else if (action == APP_LEAVE)
  {
    SEdge *e = ctx->eR;

    dbsp (">crypted_ctx_leave\n");

    while (e)
    {
      dbsp ("cleaning\n");
      if (e->next && e->next == glob_mem->crypted_msgs_tmp)
      {
	e->next = NULL;
	break;
      }
      e = e->next;
    }

    spider_clear_r (ctx);
    dbsp ("<crypted_ctx_leave\n");
  }
  return APP_OK;
}

SNodeP crypted_n = { NULL, crypted_ctx };
static SNodeP crypted_n_send = { lc_Send_crypted, crypted_send };
static SNodeP crypted_n_locked = { lc_Unlock, crypted_unlock };
static SNodeP crypted_n_unlocked = { lc_Lock, crypted_lock };
static SNodeP crypted_n_keys = { lc_Keys, crypted_keys };
static SNodeP crypted_n_keys_new = { lc_New_Key, crypted_keys_new };
static SNodeP crypted_n_keys_view = { LC_VIEW, crypted_keys_view };
static SNodeP crypted_n_keys_delete = { LC_DELETE, crypted_keys_delete };
static SNodeP crypted_n_keys_edit = { LC_EDIT, crypted_keys_edit };
static SNodeP crypted_n_keys_main = { lc_Key_Main, crypted_keys_main };
static SNodeP crypted_n_msg_view = { LC_VIEW, crypted_msg_view };
static SNodeP crypted_n_msg_reply = { lc_Msg_Reply, crypted_msg_reply };

SEdgeP crypted_edges_p[] = {
  {&crypted_n, &crypted_n_send},
  {&crypted_n_keys, &crypted_n_keys_main},
  {&crypted_n_keys, &crypted_n_keys_new},
  NULL
};

#endif
