/*
 * masked.c - part of procsms application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "conf.h"
#include "masked.h"
#include "procsms.h"

#ifdef CONFIG_MASKED

/* *INDENT-OFF* */

lc_char PROGMEM lc_Masked[]={
        LC_EN("SMaSks")
        LC_CZ("SMaSky")
        LC_END
};

static lc_char PROGMEM lc_Send_masked[]={
        LC_EN("Send Masked")
        LC_CZ("Poslat mask.")
        LC_END
};

static lc_char PROGMEM lc_Lock[]={
        LC_EN("Lock")
        LC_CZ("Zamknout")
        LC_END
};

static lc_char PROGMEM lc_Unlock[]={
        LC_EN("Unlock")
        LC_CZ("Odemknout")
        LC_END
};

static lc_char PROGMEM lc_Masks[]={
        LC_EN("Masks")
        LC_CZ("Masky")
        LC_END
};

static lc_char PROGMEM lc_Masks_New[]={
        LC_EN("New Mask")
        LC_CZ("Nova maska")
        LC_END
};

static lc_char PROGMEM lc_New_MSISDN[]={
        LC_EN("New number")
        LC_CZ("Nove cislo")
        LC_END
};

lc_char PROGMEM lc_Default_Mask[]={
        LC_EN("Masked Message")
        LC_CZ("Maskovana zprava")
        LC_END
};

static lc_char PROGMEM lc_Mask_Sender[]={
        LC_EN("Mask Sender:")
        LC_CZ("Maskovaci odesilatel:")
        LC_END
};

static lc_char PROGMEM lc_Mask_Text[]={
        LC_EN("Mask Message:")
        LC_CZ("Maskovaci zprava:")
        LC_END
};

/* *INDENT-ON* */

Mask_key *new_masked_key ();
u8 delete_masked_key ();

void substitute_mask (u8 * s, u8 * text, u8 * sender)
{
  u8 i;
  u8 j;
  u8 *t;
  u8 len;
  u8 *tmp_sender;
  u8 text_src = 0;
  u8 sender_src = 0;

  dbsp (">substitute_mask\n");

#ifdef DEBUG
  for (i = 0; i < 0xB0; i++)
  {
    dbch (s[i]);
    dbc (' ');
  }
  dbc ('\n');
#endif

  i = strlen (text);
  dbsp ("LEN ");
  dbch (i);
  dbc (' ');
  dbih (text);
  dbc (' ');
  dbc ('\n');

  if (rb (text) == STRING_SLV)
    text_src = STRING_SLV;

  t = tpdu_seek (s, T_SMS_UD);
  if (text_src == STRING_SLV)
    text += 3;
  memcpy (t, text, i);

  if (text_src != STRING_SLV)
    dcs_78 (t, i, DCS_8_TO_7);
  *(t - 1) = i;

#ifdef MASK_SENDER
  if (rb (sender) == STRING_SLV)
    sender_src = STRING_SLV;
  t = tpdu_seek (s, T_SMS_AA);
  i = *t / 2;
  if (*t & 1)
    i++;

  len = strlen (sender);
  if (sender_src == STRING_SLV)
    sender += 3;
  tmp_sender = malloc (len);
  memcpy (tmp_sender, sender, len);
  if (sender_src != STRING_SLV)
    len = dcs_78 (tmp_sender, len, DCS_8_TO_7);

  if (i < len)
  {
    dbsp ("SENDER_< ");
    dbch (len);
    dbc (' ');
    dbch (i);
    dbc ('\n');
    for (j = 0xB0; j > 0; j--)
      t[1 + i + j] = t[1 + len + j];
  }
  else if (i > len)
  {
    dbsp ("SENDER_> ");
    dbch (len);
    dbc (' ');
    dbch (i);
    dbc ('\n');
    for (j = 0; j < 0xB0; j++)	//sms record len max.
      t[2 + len + j] = t[2 + i + j];
  }

  memcpy (t + 2, tmp_sender, len);

  *t = len;
  *(t + 1) = 0xD0;

  free (tmp_sender);
#endif

#ifdef DEBUG
  for (i = 0; i < 0xB0; i++)
  {
    dbch (s[i]);
    dbc (' ');
  }
  dbc ('\n');
#endif

  dbsp ("<substitute_mask\n");
}

u8 masked_send (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    return send_msg (MSG_MASKED, NULL, NULL);
  }
  return APP_OK;
}

u8 masked_lock (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    free (glob_mem->masked_new_key);
    glob_mem->masked_new_key = NULL;
    glob_mem->masked_key = NULL;
    glob_mem->mask_locked = 1;
  }
  return APP_OK;
}

u8 masked_unlock (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res;
    u8 k_tmp[16];
    Mask_key *key = glob_mem->mask_keys;
    u8 i;

    i = password (k_tmp, 16, NULL);
    if (i != APP_OK)
      return i;

    glob_mem->masked_key = NULL;
    while (key != NULL)
    {
      u8 *k = &key->key;

      dbsp ("masked_unlock\n");

      for (i = 0; i < 16; i++)
	if (rb (k + i) != k_tmp[i])
	  goto X1;

      dbsp ("masked_unlock_found\n");
      glob_mem->masked_key = key;
      break;
    X1:
      key = rw (&key->next);
    }

    if (key == NULL)
    {
      dbsp ("masked_unlock_notfound\n");
      glob_mem->masked_new_key = malloc (16);
      if (glob_mem->masked_new_key == NULL)
	return perror (ERR_NO_RAM);
      memcpy (glob_mem->masked_new_key, k_tmp, 16);
    }

    glob_mem->mask_locked = 0;
    return APP_BACK;
  }
  return APP_OK;
}

u8 masked_msisdn_delete (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 res;
    SNode *n = ctx->parent->f;

    res = yesno (rw (&n->text), Q_YESNO_DELETE_QMARK, 0);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES)
    {
      SEdge *e = rw (&glob_mem->masked_key->mask_msisdns);
      SEdge *f = e;

      while (e)
      {
	if (rw (&e->t) == n)
	  break;
	f = e;
	e = rw (&e->next);
      }

      if (f == e)
      {
	ww (&glob_mem->masked_key->mask_msisdns, rw (&e->next));
      }
      else
      {
	ww (&f->next, rw (&e->next));
      }

      efree (e);
      efree (rw (&n->text));
      efree (rw (&n->p));
      efree (n);

      delete_masked_key ();

      spider_back (ctx);
    }
    return APP_BACK;
  }
  return APP_OK;
}

u8 masked_msisdn_view (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    SNode *n = ctx->parent->f;
    u8 *ms = msisdn2str (rw (&n->p), MSISDN_ADN, MEM_R);
    u8 i;

    i = display_text (ms, rw (&n->text));
    free (ms);
    return i;
  }
  return APP_OK;
}

extern SNodeP masked_n_msisdn_delete;
extern SNodeP masked_n_msisdn_view;
extern SNodeP masked_n;

u8 masked_msisdn (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    spider_append_r (ctx, &masked_n_msisdn_delete);
    spider_append_r (ctx, &masked_n_msisdn_view);
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
  }
  return APP_OK;
}

extern SNodeP masked_n_masks_view;
extern SNodeP masked_n_masks_edit;
extern SNodeP masked_n_masks_delete;
extern SNodeP masked_n_masks;

u8 masked_masks_ctx (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    spider_append_r (ctx, &masked_n_masks_delete);
    spider_append_r (ctx, &masked_n_masks_edit);
    spider_append_r (ctx, &masked_n_masks_view);
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
  }
  return APP_OK;
}

u8 masked_masks_new (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    Mask *mask = NULL;
    u8 *mask_sender = NULL;
    u8 *mask_text = NULL;
    SNode *n = NULL;
    u8 *n_text = NULL;
    SEdge *e = NULL;
    u8 *res;
    u8 i, j;
    Mask_key *key = glob_mem->masked_key;

    mask = emalloc (sizeof (Mask));
    if (mask == NULL)
      goto X_ERR;

    n = emalloc (sizeof (SNode));
    if (n == NULL)
      goto X_ERR;

    e = emalloc (sizeof (SEdge));
    if (e == NULL)
      goto X_ERR;

#ifdef MASKS_SENDER
    res =
      get_input (locale (lc_Mask_Sender), 1, 0x0F, NULL,
		 Q_GET_INPUT_ALPHABET);
    if (res == ENULL)
    {
      i = APP_END;
      goto X_END;
    }
    if (res == NULL)
    {
      i = APP_BACK;
      goto X_END;
    }

    i = res[0] - 1;		//len-dcs
    mask_sender = emalloc (i + 1);
    if (mask_sender == NULL)
      goto X_ERR;
    memcpy (mask_sender, res + 2, i);
    wb (mask_sender + i, '\0');
#endif

    res =
      get_input (locale (lc_Mask_Text), 1, 160, NULL, Q_GET_INPUT_ALPHABET);
    if (res == ENULL)
    {
      i = APP_END;
      goto X_END;
    }
    if (res == NULL)
    {
      i = APP_BACK;
      goto X_END;
    }

    i = res[0] - 1;
    if (i > 10)
      i = 10;
    n_text = emalloc (i + 1);
    if (n_text == NULL)
      goto X_ERR;
    memcpy (n_text, res + 2, i);
    wb (n_text + i, '\0');

    i = res[0] - 1;
    res[1] = i;
    i = dcs_78 (res + 2, i, DCS_8_TO_7) + 2;
    res[0] = 0;			//7b dcs
    mask_text = emalloc (i + 1);	// STRING_SLV, len, dcs, text_7b
    if (mask_text == NULL)
      goto X_ERR;

    wb (mask_text, STRING_SLV);
    memcpy (mask_text + 1, res, i);

    if (key == NULL)
    {
      key = new_masked_key ();
      if (key == NULL)
	goto X_ERR;
    }

    dbsp ("MASKED_NEW_MASK\n");

    ww (&mask->sender, mask_sender);
    ww (&mask->text, mask_text);

    ww (&n->text, n_text);
    ww (&n->cb, masked_masks_ctx);
    ww (&n->p, mask);

    ww (&e->f, &masked_n_masks);
    ww (&e->t, n);
    wb (&e->source, MEM_F_P | MEM_T_E);

    ww (&e->next, rw (&key->mask_masks));
    ww (&key->mask_masks, e);

    wb (&key->nr_masks, rb (&key->nr_masks) + 1);

    return APP_BACK;
  X_ERR:
    i = APP_OK;
  X_END:
    efree (mask);
    efree (mask_sender);
    efree (mask_text);
    efree (e);
    efree (n);
    efree (n_text);

    if (i == APP_OK)
      perror (ERR_NO_EEPROM);
    return i;
  }
  return APP_OK;
}

u8 masked_masks_view (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    SNode *n = ctx->parent->f;
    Mask *m = rw (&n->p);

    dbsp ("masks_view ");
    dbih (rw (&m->text));
    dbc ('\n');
    return display_text (rw (&m->text), rw (&m->sender));	// m->text SLV
  }
  return APP_OK;
}

u8 masked_masks_edit (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res;
    SNode *n = ctx->parent->f;
    Mask *m = rw (&n->p);
    u8 *x, *y;
    u8 i;

    x = y = NULL;

#ifdef MASK_SENDER
    res =
      get_input (locale (lc_Mask_Sender), 1, 0x0F, rw (&m->sender),
		 Q_GET_INPUT_ALPHABET);
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;

    i = res[0] - 1;		//len-dcs
    x = emalloc (i + 1);
    if (x == NULL)
      goto X_ERR;
    memcpy (x, res + 2, i);
    wb (x + i, '\0');
    efree (rw (&m->sender));
    ww (&m->sender, x);
#endif

    res = get_input (locale (lc_Mask_Text), 1, 160, rw (&m->text), Q_GET_INPUT_ALPHABET);	//m->text MEM_SLV
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;

    i = res[0] - 1;
    res[1] = i;
    i = dcs_78 (res + 2, i, DCS_8_TO_7) + 2;
    res[0] = 0;			//7b dcs
    y = emalloc (i + 1);	// len, dcs, text_7b
    if (y == NULL)
      goto X_ERR;

    wb (y, STRING_SLV);
    memcpy (y + 1, res, i);

    efree (rw (&m->text));
    ww (&m->text, y);

    return APP_BACK;
  X_ERR:
    efree (x);
    efree (y);
    return perror (ERR_NO_EEPROM);
  }
  return APP_OK;
}

u8 masked_masks_delete (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 res;
    SNode *n = ctx->parent->f;

    res = yesno (rw (&n->text), Q_YESNO_DELETE_QMARK, 0);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES)
    {
      SEdge *e = rw (&glob_mem->masked_key->mask_masks);
      SEdge *f = e;
      Mask *mask = rw (&n->p);

      while (e)
      {
	if (rw (&e->t) == n)
	  break;
	f = e;
	e = rw (&e->next);
      }

      if (f == e)
      {
	ww (&glob_mem->masked_key->mask_masks, rw (&e->next));
      }
      else
      {
	ww (&f->next, rw (&e->next));
      }

      efree (e);
      efree (rw (&n->text));
      efree (rw (&mask->sender));
      efree (rw (&mask->text));
      efree (rw (&n->p));
      efree (n);

      delete_masked_key ();

      spider_back (ctx);
    }
    return APP_BACK;
  }
  return APP_OK;
}

u8 masked_masks (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    if (glob_mem->masked_key)
      ctx->eE = rw (&glob_mem->masked_key->mask_masks);
  }
  else if (action == APP_LEAVE)
  {
    ctx->eE = NULL;
  }
  return APP_OK;
}

u8 delete_masked_key ()
{
  Mask_key *key = glob_mem->masked_key;
  Mask_key *k, *l;
  u8 *new_key;
  u8 new_key_len;

  if (rw (&key->mask_masks) != NULL || rw (&key->mask_msisdns) != NULL)
    return NO_ERROR;

  new_key = malloc (16);
  if (new_key == NULL)
    return ERR_NO_RAM;
  memcpy (new_key, &key->key, 16);

  glob_mem->masked_new_key = new_key;

  k = l = glob_mem->mask_keys;
  while (k)
  {
    if (k == key)
      break;
    l = k;
    k = rw (&k->next);
  }

  if (k = l)
  {
    l = rw (&k->next);
    glob_mem->mask_keys = l;
    {
      Pers_mem *p = app_data ();

      ww (&p->mask_keys, l);
    }
  }
  else
  {
    ww (&l->next, rw (&k->next));
  }

  efree (key);

  return NO_ERROR;
}

Mask_key *new_masked_key ()
{
  Mask_key *key = emalloc (sizeof (Mask_key));

  if (key != NULL)
  {
    dbsp (">NEW_MASK_KEY\n");
    memcpy (&key->key, glob_mem->masked_new_key, 16);
    ww (&key->mask_msisdns, NULL);
    ww (&key->mask_masks, NULL);
    wb (&key->nr_masks, 0);
    ww (&key->next, glob_mem->mask_keys);
    {
      Pers_mem *p = app_data ();

      ww (&p->mask_keys, key);
      glob_mem->mask_keys = key;
    }
    glob_mem->masked_key = key;

    free (glob_mem->masked_new_key);
    glob_mem->masked_new_key = NULL;

    dbsp ("<NEW_MASK_KEY\n");
  }
  return key;
}

u8 masked_msisdn_new (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res;
    SEdge *e = NULL;
    SEdge *f, *g;
    SNode *n = NULL;
    u8 *ealpha = NULL;
    u8 alpha[ALPHA_LEN];
    u8 alpha_len;

    db_malloc ();

    res = msisdn (alpha);

    db_malloc ();
    if (res == ENULL)
      return APP_END;
    if (res != NULL)
    {
      res = msisdncpy (res, MSISDN_ADN, MEM_E);
      db_malloc ();
      if (res != NULL)
      {
	Mask_key *key = glob_mem->masked_key;

	e = emalloc (sizeof (SEdge));
	if (e == NULL)
	  goto X_ERR;

	n = emalloc (sizeof (SNode));
	if (n == NULL)
	  goto X_ERR;

	alpha_len = strlen (alpha) + 1;
	ealpha = emalloc (alpha_len);
	if (ealpha == NULL)
	  goto X_ERR;

	memcpy (ealpha, alpha, alpha_len);

	if (key == NULL)
	{
	  key = new_masked_key ();
	  if (key == NULL)
	    goto X_ERR;
	}

	dbsp ("MASKED_NEW_MSISDN\n");
	db_malloc ();

	ww (&n->text, ealpha);
	ww (&n->cb, masked_msisdn);
	ww (&n->p, res);

	ww (&e->f, &masked_n);
	ww (&e->t, n);
	wb (&e->source, MEM_F_P | MEM_T_E);

	insert_sorted_alpha (e, &key->mask_msisdns, alpha);
      }
      else
	goto X_ERR;
    }

    db_malloc ();
    return APP_BACK;
  X_ERR:
    efree (res);
    efree (e);
    efree (ealpha);
    efree (n);

    return perror (ERR_NO_EEPROM);
  }
  return APP_OK;
}

extern SNodeP masked_n_locked;
extern SNodeP masked_n_unlocked;
extern SNodeP masked_n_masks;
extern SNodeP masked_n_msisdn_new;

u8 masked_ctx (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    if (glob_mem->mask_locked == 0)
    {
      if (glob_mem->masked_key)
	ctx->eE = rw (&glob_mem->masked_key->mask_msisdns);

      spider_append_r (ctx, &masked_n_msisdn_new);
      spider_append_r (ctx, &masked_n_masks);
      spider_append_r (ctx, &masked_n_unlocked);
      ctx->n_default = 2;
    }
    else
      spider_append_r (ctx, &masked_n_locked);
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
    ctx->eE = NULL;
  }
  return APP_OK;
}

SNodeP masked_n = { NULL, masked_ctx };
static SNodeP masked_n_send = { lc_Send_masked, masked_send };
static SNodeP masked_n_locked = { lc_Unlock, masked_unlock };
static SNodeP masked_n_unlocked = { lc_Lock, masked_lock };
static SNodeP masked_n_masks = { lc_Masks, masked_masks };
static SNodeP masked_n_masks_new = { lc_Masks_New, masked_masks_new };
static SNodeP masked_n_masks_view = { LC_VIEW, masked_masks_view };
static SNodeP masked_n_masks_edit = { LC_EDIT, masked_masks_edit };
static SNodeP masked_n_masks_delete = { LC_DELETE, masked_masks_delete };
static SNodeP masked_n_msisdn_new = { lc_New_MSISDN, masked_msisdn_new };
static SNodeP masked_n_msisdn_delete = { LC_DELETE, masked_msisdn_delete };
static SNodeP masked_n_msisdn_view = { LC_VIEW, masked_msisdn_view };

SEdgeP masked_edges_p[] = {
  {&masked_n, &masked_n_send},
  {&masked_n_masks, &masked_n_masks_new},
  NULL
};

#endif
