/*
 * procsms.c - part of procsms application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "conf.h"
#include "procsms.h"
#include "archive.h"
#include "crypted.h"
#include "masked.h"

/*
TODO:

- optimize sms_rec - reread only modified(=masked or crypted)
- APDU rec_nr not direct
- ?mask all
- ?better view crypted (show head)
*/

/* *INDENT-OFF* */

static lc_char PROGMEM lc_Enter_Text[]={
        LC_EN("Enter Text:")
        LC_CZ("Text zpravy:")
        LC_END
};

/*
TODO
static lc_char PROGMEM lc_Send[]={
        LC_EN("Send?")
        LC_CZ("Poslat?")
        LC_END
};
*/

lc_char PROGMEM lc_Msg_Reply[]={
        LC_EN("Reply")
        LC_CZ("Odpovedet")
        LC_END
};

lc_char PROGMEM lc_Msg_Forward[]={
        LC_EN("Forward")
        LC_CZ("Preposlat")
        LC_END
};

static lc_char PROGMEM lc_Sending_SMS[]={
        LC_EN("Sending Message")
        LC_CZ("Posilam zpravu")
        LC_END
};

static lc_char PROGMEM lc_SMS_Send_OK[]={
        LC_EN("Message Sent")
        LC_CZ("Zprava odeslana")
        LC_END
};

static lc_char PROGMEM lc_SMS_Send_Not_OK[]={
        LC_EN("Sending was not successful")
        LC_CZ("Zpravu se nepodarilo odeslat")
        LC_END
};


#ifdef CONFIG_ONDISPLAY
static lc_char PROGMEM lc_On_Display[] = {
  	LC_EN ("On Display") 
	LC_CZ ("Na displej") 
	LC_END
};
#endif

/* *INDENT-ON* */

#ifdef CONFIG_CRYPTED
static u8 PROGMEM def_main_key[] = { 'a', 'd', 'g', 'j', 'm', 'p', 't', 'w' };
static u8 PROGMEM crypted_head[] = { 0x45, 0xe7, 0x50, 0x9a, 0x85, 0x52, 0x8b, 0x44, 0x97, 0xcb, 0xd5, 0x28 };	//="ENCRYPTED...\r" in 7b"
#endif

#ifdef CONFIG_MASKED
static u8 PROGMEM t_Default_Mask_Sender[] = { "1189" };
#endif

u8 PROGMEM ef_sms_path[] = { 0x3f, 0x00, 0x7f, 0x10, 0x6f, 0x3c };

#define DEF_SMS_REC_NR		0x0A
Smsing_mem *glob_mem;

void insert_sorted_alpha (SEdge * e, SEdge ** start, u8 * alpha)
{
//adn sorted
  u8 i, len;
  u8 *t;
  SEdge *f, *g;
  SNode *n;

  dbsp (">insert_sorted_alpha\n");

  g = f = rw (start);
  while (f)
  {
    dbsp ("ADN insert sorted\n");
    n = rw (&f->t);
    t = rw (&n->text);
    len = strlen (t);
    dbsp ("L ");
    dbch (len);
    dbc (' ');
    dbch (alpha[1]);
    dbc ('\n');
    if (len > alpha[0])
      len = alpha[0];
    for (i = 0; i < len; i++)
    {
      dbch (alpha[1 + i]);
      dbc (' ');
      dbch (rb (t + i));
      dbc ('\n');
      if (tolower (alpha[1 + i]) > tolower (rb (t + i)))
	goto X_NEXT;
    }
    dbsp ("X_BREAK\n");
    break;
  X_NEXT:
    dbsp ("X_NEXT\n");
    g = f;
    f = rw (&f->next);
  }

  if (f == g)
  {
    dbsp ("F==G\n");
    ww (start, e);
  }
  else
  {
    dbsp ("F!=G\n");
    ww (&g->next, e);
  }

  ww (&e->next, f);
  dbsp ("<insert_sorted_alpha\n");
}

u8 check_crypted (u8 * s)
{
  u8 i;
  u8 *t;

  t = tpdu_seek (s, T_SMS_UDL);
  if (*t != 0xA0)
    return 0;
  t++;
  /*
     for (i = 0; i < sizeof (crypted_head); i++)
     if (rb (crypted_head + i) != t[i])
     return 0;
   */

  return (memcmp (t, crypted_head, sizeof (crypted_head)) == 0) ? 1 : 0;
}

u8 check_mask (u8 * s, u8 p1)
{
  // p1==0xff -> archive
  Mask_key *key = glob_mem->mask_keys;
  u8 *t;
  SEdge *msisdns;
  SNode *n;
  u8 i;
  Crypted_msg *msg;

  t = tpdu_seek (s, T_SMS_AA);
  while (key)
  {
    if (glob_mem->mask_locked == 1 || key != glob_mem->masked_key)
    {
      msisdns = rw (&key->mask_msisdns);
      while (msisdns)
      {
	n = rw (&msisdns->t);
	if (msisdncmp (t, MSISDN_SMS, rw (&n->p), MSISDN_ADN))
	{
	  u8 nr_masks = rb (&key->nr_masks);

	  dbsp ("MASK_FOUND\n");

	  if (p1 != 0xFF)
	  {
	    // delete from crypted list

	    msisdns = glob_mem->crypted_msgs;
	    while (msisdns)
	    {
	      n = msisdns->t;
	      msg = n->p;
	      if (msg->p1 == p1)
		goto X_HAS_CRYPTED;
	      msisdns = msisdns->next;
	    }
	    goto X_HAS_NOT_CRYPTED;

	  X_HAS_CRYPTED:
	    {
	      SEdge *msisdns = glob_mem->crypted_msgs;
	      SEdge *e;
	      SNode *n;
	      Crypted_msg *msg;

	      e = msisdns;
	      while (msisdns)
	      {
		n = msisdns->t;
		msg = n->p;
		if (msg->p1 == p1)
		{
		  dbsp ("delete crypted msg\n");
		  free (msg->msisdn);
		  free (n->text);
		  free (n);

		  // delete crypted msg
		  if (e == msisdns)
		  {
		    glob_mem->crypted_msgs = e->next;
		  }
		  else
		  {
		    e->next = msisdns->next;
		  }
		  free (msisdns);
		  break;
		}

		e = msisdns;
		msisdns = msisdns->next;
	      }

	    }

	  X_HAS_NOT_CRYPTED:
	  }

	  if (nr_masks == 0)
	    substitute_mask (s, locale (lc_Default_Mask),
			     t_Default_Mask_Sender);
	  else
	  {
	    SEdge *m_e = rw (&key->mask_masks);
	    SNode *m_n;
	    Mask *mask;

	    i = p1;
	    if (nr_masks < i)
	      i = i % nr_masks;
	    else
	      i = nr_masks % i;

	    dbsp ("RND_MASK ");
	    dbch (nr_masks);
	    dbc (' ');
	    dbch (p1);
	    dbc (' ');
	    dbch (i);
	    dbc (' ');
	    dbc ('\n');

	    while (i)
	    {
	      m_e = rw (&m_e->next);
	      i--;
	    }
	    m_n = rw (&m_e->t);
	    mask = rw (&m_n->p);

	    substitute_mask (s, rw (&mask->text), rw (&mask->sender));	// mask->text MEM_SLV
	  }
	  goto X1;
	}
	msisdns = rw (&msisdns->next);
      }
    }

    key = rw (&key->next);
  }
  return 0;
X1:
  return 1;
}

void fake_sms_file (File_apdu_data * fa)
{
  u8 i;

  dbsp ("FAKE_SMS_FILE\n");

  if (fa->ins == ME_CMD_GET_RESPONSE)
  {
    u16 nr_sms;
    u8 i;

    dbsp ("FAKE_GET_RESPONSE\n");
    sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
    nr_sms = ((fa->data[2] << 8) | fa->data[3]) / fa->data[14];
    dbsp ("#SMS ");
    dbih (nr_sms);
    dbc ('\n');

    dbsp ("SMS_REC REALLOC\n");
    free (glob_mem->sms_rec);
    // what if get resp more often???
    glob_mem->sms_rec = malloc (nr_sms);
    glob_mem->sms_rec_nr = nr_sms;

    return;
  }

  if (fa->ins == ME_CMD_READ_RECORD)
  {
    SEdge *msisdns;
    SNode *n;
    u8 *t, *s;
    u8 i, len;
    Crypt_key *c_key;
    Crypted_msg *msg;

    dbsp ("FAKE_READ_RECORD\n");
    sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
    s = fa->data;
    if (fa->p1 > glob_mem->sms_rec_nr)
    {
      dbsp ("SMS_REC REALLOC\n");
      free (glob_mem->sms_rec);
      glob_mem->sms_rec = malloc (fa->p1);
      glob_mem->sms_rec_nr = fa->p1;
      for (i = 0; i < fa->p1; i++)
	glob_mem->sms_rec[i] = 0xFF;
    }
    glob_mem->sms_rec[fa->p1 - 1] = *s;
    dbsp ("P ");
    dbch (glob_mem->sms_rec[fa->p1 - 1]);
    dbc ('\n');
    s++;
    i = *s;
    s = s + 1 + i;		//skip SMSC

    if (check_mask (s, fa->p1))
      return;

    // Not masked

    dbsp ("CR_0\n");
    if (check_crypted (s) == 0)
      return;

    dbsp ("CR_2\n");

    msisdns = glob_mem->crypted_msgs;
    while (msisdns)
    {
      n = msisdns->t;
      msg = n->p;
      if (msg->p1 == fa->p1)
      {
	if (msg->adn_ready == 0 && adn_ready ())
	{
	  u8 *tmp_ms;

	  // update msg->msisdn
	  free (n->text);
	  t = tpdu_seek (s, T_SMS_AA);
	  tmp_ms = msisdncpy (t, MSISDN_SMS, MEM_R);
	  n->text = get_adn (tmp_ms, MSISDN_SMS);
	  free (tmp_ms);
	  msg->adn_ready = 1;

	  sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
	}
	goto X2;
      }
      msisdns = msisdns->next;
    }

    {
      u8 *ms;
      u8 *ms_nr;
      SEdge *f, *g;
      Crypted_msg *m;

      dbsp ("create crypted item\n");

      n = malloc (sizeof (SNode));
      if (n == NULL)
	goto X1;
      msisdns = malloc (sizeof (SEdge));
      if (msisdns == NULL)
      {
	free (n);
	goto X1;
      }
      msg = malloc (sizeof (Crypted_msg));
      if (msg == NULL)
      {
	free (n);
	free (msisdns);
	goto X1;
      }

      t = tpdu_seek (s, T_SMS_AA);
      ms_nr = malloc (*t);
      memcpy (ms_nr, t, *t);
      ms = get_adn (ms_nr, MSISDN_SMS);
#ifdef DEBUG
      {
	u8 ik;
	u8 il = strlen (ms);

	dbsp ("MS_X ");
	for (ik = 0; ik < il; ik++)
	  dbc (ms[ik]);
	dbc ('\n');
      }
#endif
      sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);

      if (ms == NULL)
      {
	free (msisdns);
	free (n);
	free (msg);
	free (ms_nr);
	goto X1;
      }

      msg->p1 = fa->p1;
      msg->msisdn = ms_nr;
      msg->adn_ready = adn_ready ();
      t = tpdu_seek (s, T_SMS_SCTS);
      memcpy (&msg->time, t, 6);

      n->text = ms;
      n->cb = crypted_msg_ctx;
      n->p = msg;

      msisdns->f = &crypted_n;
      msisdns->t = n;
      msisdns->source = MEM_F_P | MEM_T_R;

      //insert time sorted

      g = f = glob_mem->crypted_msgs;
      while (f)
      {
	dbsp ("insert time sorted\n");

	n = f->t;
	m = n->p;
	for (i = 0; i < 6; i++)
	{
	  dbsp ("C: ");
	  dbch (swap_nibble (msg->time[i]));
	  dbc (' ');
	  dbch (swap_nibble (m->time[i]));
	  dbc ('\n');
	  if (swap_nibble (msg->time[i]) > swap_nibble (m->time[i]))
	    goto X_PREV;
	  if (swap_nibble (msg->time[i]) < swap_nibble (m->time[i]))
	    break;
	}
	g = f;
	f = f->next;
      }
    X_PREV:
      if (f == g)
      {
	glob_mem->crypted_msgs = msisdns;
      }
      else
      {
	g->next = msisdns;
      }

      msisdns->next = f;
    }

  X2:
#ifdef DECRYPT_ON_FLY
    if (glob_mem->crypt_key != NULL)
    {
      dbsp ("CR_5\n");

      t = tpdu_seek (s, T_SMS_AA);

      msisdns = glob_mem->crypted_keys;
      while (msisdns)
      {
	dbsp ("CR_4\n");
	n = rw (&msisdns->t);
	c_key = rw (&n->p);
	if (msisdncmp (t, MSISDN_SMS, rw (&c_key->msisdn), MSISDN_ADN))
	{
	  dbsp ("CR_5\n");
	  // found crypted key
	  u8 *in = malloc (16);
	  Twofish_key *ak = malloc (sizeof (Twofish_key));

	  if (buf && ak)
	  {
	    u8 *msg;
	    u8 j;

	    t = tpdu_seek (s, T_SMS_UD);
	    msg = t + 12;

	    twofish_ecb_decrypt (&c_key->text, in, glob_mem->crypt_key);
	    twofish_setup (in, 16, ak);
	    twofish_cbc_decrypt (msg, msg, 8, ak);

	    // decrypt message
	    /*
	       for (j = 0; j < 8; j++)
	       {
	       twofish_ecb_decrypt (msg + j * 16, buf, ak);
	       if (j)
	       for (i = 0; i < 16; i++)
	       buf[i] = buf[i] ^ in[i];
	       memcpy (in, msg + j * 16, 16);
	       memcpy (msg + j * 16, buf, 16);
	       }
	     */
	    len = t[20];
	    if (len > 0xA0)
	      len = 0xA0;

	    for (i = 0; i < len; i++)
	      t[i] = msg[i + 9];
	  }
	  *(t - 1) = dcs_7b_len (len);

	  free (ak);
	  free (in);
	  goto X1;
	}

	msisdns = rw (&msisdns->next);
      }
    }
#endif

    dbsp ("CR_3\n");
    t = tpdu_seek (s, T_SMS_UD);
    i = strlen (locale (lc_Default_Crypted));
    memcpy (t, locale (lc_Default_Crypted), i);
    dcs_78 (t, i, DCS_8_TO_7);
    *(t - 1) = i;

  X1:
    return;
  }

  if (fa->ins == ME_CMD_UPDATE_RECORD)
  {
    u8 *s = fa->data;
    u8 i = *s;

    dbsp ("FAKE_UPDATE_RECORD\n");

    if (fa->p1 > glob_mem->sms_rec_nr)
    {
      dbsp ("SMS_REC REALLOC\n");
      free (glob_mem->sms_rec);
      glob_mem->sms_rec = malloc (fa->p1);
      glob_mem->sms_rec_nr = fa->p1;
      for (i = 0; i < fa->p1; i++)
	glob_mem->sms_rec[i] = 0xFF;
    }

    dbsp ("I ");
    dbch (i);
    dbc ('\n');
    dbsp ("P ");
    dbch (glob_mem->sms_rec[fa->p1 - 1]);
    dbc ('\n');

    if (glob_mem->sms_rec[fa->p1 - 1] != i && i != 0x03
	&& glob_mem->sms_rec[fa->p1 - 1] != 0x00 && i != 0)
    {
      dbsp ("FAKE_UPDATE_RECORD_READ\n");
      glob_mem->sms_rec[fa->p1 - 1] = *s;
      sim (ME_CMD_READ_RECORD, fa->p1, fa->p2, fa->p3, fa->data);
      *s = glob_mem->sms_rec[fa->p1 - 1];
    }

    if (i == 0x00)
    {
      SEdge *msisdns = glob_mem->crypted_msgs;
      SEdge *e;
      SNode *n;
      Crypted_msg *msg;

      e = msisdns;
      while (msisdns)
      {
	n = msisdns->t;
	msg = n->p;
	if (msg->p1 == fa->p1)
	{
	  dbsp ("delete crypted msg\n");
	  free (msg->msisdn);
	  free (n->text);
	  free (n);

	  // delete crypted msg
	  if (e == msisdns)
	  {
	    glob_mem->crypted_msgs = e->next;
	  }
	  else
	  {
	    e->next = msisdns->next;
	  }
	  free (msisdns);
	  break;
	}

	e = msisdns;
	msisdns = msisdns->next;
      }

    }

    glob_mem->sms_rec[fa->p1 - 1] = i;

#ifdef CONFIG_ARCHIVE
    archive_store_sms (fa->data, SMS_PLAIN, fa->p1, fa->p2);
#else
    sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
#endif
    fa->data[0] = 0x90;
    fa->data[1] = 0x00;
    return;
  }

  dbsp ("FAKE_OTHER_CMD\n");
  sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
  return;
}

u8 send_msg (u8 type, u8 * def_text, u8 * def_ms)	// def_ms is MSISDN_SMS
{
  u8 *res;
  u8 *ms = NULL;
  u8 *text = def_text;
  u8 i;
  u8 len = 160;
  u8 err;
  u8 dcs = 0x12;
  u8 *buf = buf_B ();
  u8 *tmp;
  Twofish_key *ak = NULL;

  if (type == MSG_ON_DISPLAY)
    dcs = 0xF0;

X_TEXT:
  if (type == MSG_CRYPTED)
    len = 136;

  res = get_input (locale (lc_Enter_Text), 0, len, text, Q_GET_INPUT_ALPHABET | Q_GET_INPUT_SMS_PACKED);	//text STRING_SLV
  if (res == ENULL)
  {
    i = APP_END;
    goto X_END;
  }
  if (res == NULL)
  {
    i = APP_BACK;
    goto X_END;
  }

  if (text != def_text)
  {
    free (text);
    text = NULL;
  }
  len = res[0] - 1;
  text = malloc (len + 2);	//STRING_SLV+dcs+len
  if (text == NULL)
  {
    err = ERR_NO_RAM;
    goto X_ERR;
  }
  text[0] = STRING_SLV;
  text[1] = res[1];
  text[2] = len;
  memcpy (text + 3, res + 2, len);

X_MSISDN:
  if (def_ms == NULL)
  {
    res = msisdn (NULL);
    if (res == ENULL)
    {
      i = APP_END;
      goto X_END;
    }
    if (res == NULL)
    {
      i = APP_BACK;
      goto X_TEXT;
    }
    def_ms = msisdn2str (res, MSISDN_ADN, MEM_R);
    ms = str2msisdn (def_ms, MSISDN_SMS, MEM_R);
    free (def_ms);
    def_ms = NULL;
    if (ms == NULL)
    {
      err = ERR_NO_RAM;
      goto X_ERR;
    }
  }
  else
    ms = def_ms;		//reply

X_KEY:
  if (type == MSG_CRYPTED)
  {
    u8 *out = buf;
    u8 *key = out + 16;
    u8 j;

    tmp = key + 16;

    ak = malloc (sizeof (Twofish_key));
    if (ak == NULL)
    {
      err = ERR_NO_RAM;
      goto X_ERR;
    }

    if (glob_mem->crypt_key != NULL)
    {
      SEdge *msisdns;
      SNode *n;
      Crypt_key *c_key;

      msisdns = glob_mem->crypted_keys;

      while (msisdns)
      {
	n = rw (&msisdns->t);
	c_key = rw (&n->p);
	if (msisdncmp (ms, MSISDN_SMS, rw (&c_key->msisdn), MSISDN_ADN))
	{
	  twofish_ecb_decrypt (&c_key->text, key, glob_mem->crypt_key);
	  goto X1;
	}
	msisdns = rw (&msisdns->next);
      }
    }
    //ask for key to key[]

    i = password (key, 16, NULL);
    if (i != APP_OK)
      goto X_END;

  X1:
    xrand (out);
    err = twofish_setup (key, 16, ak);
    if (err != NO_ERROR)
      goto X_ERR;

    len = text[2];
    memcpy (tmp, crypted_head, sizeof (crypted_head));

    for (i = 0; i < 8; i++)
      tmp[12 + i] = out[i];	//rand
    tmp[20] = len;
    for (i = 0; i < len; i++)
      tmp[21 + i] = text[3 + i];
    for (; i < 119; i++)
      tmp[21 + i] = 140 - len;	//padding

    err = twofish_cbc_encrypt (&tmp[12], &tmp[12], 8, ak);
    if (err != NO_ERROR)
      goto X_ERR;

    free (ak);
    ak = NULL;

    len = 140;
  }
  else
  {
    len = text[2];
    memcpy (buf, text + 3, len);
    tmp = buf;
  }

X_SEND:

  res =
    send_sms (tmp, len, ms, MSISDN_SMS, dcs, 0x00, locale (lc_Sending_SMS),
	      NULL);

  i = get_tag (res, T_RESULT);
  if (i != 0)
  {
    i += 2;
    if (res[i] == 0x00)
      i = display_text (locale (lc_SMS_Send_OK), NULL);
    else
      i = display_text (locale (lc_SMS_Send_Not_OK), NULL);

    if (i == APP_OK)
    {
      i = APP_BACK;
      goto X_END;
    }
    if (i == APP_END)
      goto X_END;
    if (ms != def_ms)
    {
      free (ms);
      ms = NULL;
    }
    goto X_MSISDN;
  }

  i = APP_BACK;
  goto X_END;

X_ERR:
  i = APP_OK;

X_END:
  if (text != def_text)
    free (text);
  if (ms != def_ms)
    free (ms);
  free (ak);

  if (i == APP_OK)
    return perror (err);
  return i;
}

void smsing (Menu_selection_data * x)
{
  SCtx *c;
  u8 nr_items = 0;
  u8 it_masked = 0;
  u8 it_crypted = 0;
  u8 it_ondisplay = 0;
  u8 it_archive = 0;
  u8 item = x->item;

/*
Order:
- crypted
- masked
- ondisplay
- archive
*/

  dbsp ("ITEM :");
  dbch (item);
  dbc ('\n');

#ifdef CONFIG_CRYPTED
  it_crypted = nr_items;
#endif

#ifdef CONFIG_MASKED
  nr_items++;
  it_masked = nr_items;
#endif

#ifdef CONFIG_ONDISPLAY
  nr_items++;
  it_ondisplay = nr_items;
#endif

#ifdef CONFIG_ARCHIVE
  nr_items++;
  it_archive = nr_items;
#endif

  if (item == it_ondisplay)
  {
#ifdef CONFIG_ONDISPLAY
    dbsp ("menu_ondisplay\n");
    send_msg (MSG_ON_DISPLAY, NULL, NULL);
    return;
#endif
  }

  c = spider_init ();
  if (c == NULL)
    return;
  spider_set_order (c, ORDER_P, ORDER_R, ORDER_E, ORDER_D);

  if (item == it_crypted)
  {
#ifdef CONFIG_CRYPTED
    // crypted
    dbsp ("menu_crypted\n");
    c->n = &crypted_n;
    c->eP = &crypted_edges_p;
#endif
  }
  else if (item == it_masked)
  {
#ifdef CONFIG_MASKED
    // masked
    dbsp ("menu_masked\n");
    c->n = &masked_n;
    c->eP = &masked_edges_p;
#endif
  }
  else if (item == it_archive)
  {
#ifdef CONFIG_ARCHIVE
    // archive
    dbsp ("menu_archive\n");
    spider_set_order (c, ORDER_R, ORDER_P, ORDER_E, ORDER_D);
    seek_sms ();

    c->n = &archive_n;
    c->eP = &archive_edges_p;
    c->df_fnc = archive_df;
    glob_mem->is_rules = 0;
    glob_mem->drop = 0;
#endif
  }

  spider (c);

#ifdef CONFIG_MASKED
  free (glob_mem->masked_new_key);
  glob_mem->masked_new_key = NULL;
#endif
}

void turbo_handler (u8 action, void *data)
{
  switch (action)
  {
    case ACTION_APP_REGISTER:
      {
	Pers_mem *p = emalloc (sizeof (Pers_mem));
	u8 i;
	u8 *buf = malloc (3 * 16);
	u8 *in = buf;
	u8 *tmp = in + 16;
	u8 *key = tmp + 16;
	Twofish_key *ak = malloc (sizeof (Twofish_key));

#ifdef CONFIG_ARCHIVE
	reg_file_type (t_ext_sms, file_sms_cb);
#endif

	if (p && buf)
	{
#ifdef CONFIG_MASKED
	  ww (&p->mask_keys, NULL);
#endif

#ifdef CONFIG_ARCHIVE
	  wb (&p->activ, 0);
#endif

#ifdef CONFIG_CRYPTED
	  ww (&p->crypted_keys, NULL);
	  memcpy (tmp, def_main_key, sizeof (def_main_key));
	  make_crypt_key (key, tmp, sizeof (def_main_key));
#ifdef DEBUG
	  dbsp ("DEF_KEY: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (key[i]);
	    dbc (' ');
	  }
	  dbc ('\n');
#endif
	  xrand (in);
	  xrand (in + 8);
	  memcpy (&p->main_key_rnd, in, 16);
	  twofish_setup (key, 16, ak);
	  twofish_ecb_encrypt (in, &p->main_key_crnd, ak);
#ifdef DEBUG
	  dbsp ("KEY_RND: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (rb (&p->main_key_rnd[i]));
	    dbc (' ');
	  }
	  dbc ('\n');
	  dbsp ("KEY_CRND: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (rb (&p->main_key_crnd[i]));
	    dbc (' ');
	  }
	  dbc ('\n');

#endif
#endif
	}
	reg_app_data (p);

	free (buf);
	free (ak);
      }
      break;
    case ACTION_APP_UNREGISTER:
      {
	Pers_mem *p = app_data ();
	SEdge *e, *f;
	SNode *n;
	u8 *t;

#ifdef CONFIG_CRYPTED
	Crypted_msg *msg;
	Crypt_key *ckey;
#endif

#ifdef CONFIG_MASKED
	Mask *mask;
	Mask_key *m, *m1;
#endif

	dbsp ("UNREG crypted_keys\n");

#ifdef CONFIG_CRYPTED
	e = rw (&p->crypted_keys);
	while (e)
	{
	  n = rw (&e->t);
	  t = rw (&n->text);
	  ckey = rw (&n->p);
	  efree (rw (&ckey->msisdn));
	  efree (ckey);
	  efree (t);
	  efree (n);
	  f = rw (&e->next);
	  efree (e);
	  e = f;
	}
#endif

#ifdef CONFIG_MASKED
	dbsp ("UNREG mask_keys\n");
	m = rw (&p->mask_keys);
	while (m)
	{
	  dbsp ("UNREG mask_masks\n");
	  e = rw (&m->mask_masks);
	  while (e)
	  {
	    n = rw (&e->t);
	    t = rw (&n->text);
	    mask = rw (&n->p);
	    efree (rw (&mask->sender));
	    efree (rw (&mask->text));
	    efree (mask);
	    efree (t);
	    efree (n);
	    f = rw (&e->next);
	    efree (e);
	    e = f;
	  }
	  dbsp ("UNREG mask_msisdns\n");
	  e = rw (&m->mask_msisdns);
	  while (e)
	  {
	    n = rw (&e->t);
	    t = rw (&n->text);
	    efree (rw (&n->p));
	    efree (t);
	    efree (n);
	    f = rw (&e->next);
	    efree (e);
	    e = f;
	  }

	  m1 = rw (&m->next);
	  efree (m);
	  m = m1;
	}
#endif

	dbsp ("UNREG p\n");
	efree (p);

	if (glob_mem)
	{
	  free (glob_mem->sms_rec);
#ifdef CONFIG_CRYPTED
	  free (glob_mem->crypt_key);

	  dbsp ("UNREG crypted_msgs\n");
	  e = glob_mem->crypted_msgs;
	  while (e)
	  {
	    n = e->t;
	    t = n->text;
	    msg = n->p;
	    free (msg->msisdn);
	    free (msg);
	    free (t);
	    free (n);
	    f = e->next;
	    free (e);
	    e = f;
	  }
#endif
	  free (glob_mem);
	}
	dbsp ("UNREG_DONE\n");
      }
      break;
    case ACTION_APP_INIT:
      {
	Pers_mem *p = app_data ();
	u8 i;

	dbsp ("PROCSMS INIT\n");
	reg_file (ef_sms_path, 3);

	glob_mem = malloc (sizeof (Smsing_mem));
	if (glob_mem)
	{
	  dbsp ("PROCSMS INIT_1\n");
#ifdef CONFIG_MASKED
	  glob_mem->mask_locked = 1;
	  glob_mem->mask_keys = rw (&p->mask_keys);
	  glob_mem->masked_new_key = NULL;
#endif

#ifdef CONFIG_CRYPTED
	  glob_mem->crypt_key = NULL;
	  glob_mem->crypted_keys = rw (&p->crypted_keys);
	  glob_mem->crypted_msgs = NULL;
#endif

#ifdef CONFIG_ARCHIVE
	  glob_mem->activ = rb (&p->activ);
#endif

	  glob_mem->sms_rec = malloc (DEF_SMS_REC_NR);
	  glob_mem->sms_rec_nr = DEF_SMS_REC_NR;
	  if (glob_mem->sms_rec)
	    for (i = 0; i < DEF_SMS_REC_NR; i++)
	      glob_mem->sms_rec[i] = 0xFF;
	}

      }
      break;
    case ACTION_INSERT_MENU:
#ifdef CONFIG_CRYPTED
      insert_menu (locale (lc_Crypted));
#endif

#ifdef CONFIG_MASKED
      insert_menu (locale (lc_Masked));
#endif

#ifdef CONFIG_ONDISPLAY
      insert_menu (locale (lc_On_Display));
#endif

#ifdef CONFIG_ARCHIVE
      if (mc_detect () == MC_OK)
	insert_menu (locale (lc_Archive));
#endif
      break;
    case ACTION_MENU_SELECTION:
      stk_thread (smsing, data);
      break;
    case ACTION_FILE_APDU:
      fake_sms_file (data);
      break;
    default:
      break;
  }
}
